"""
Content controls the decomposition of python objects into their component parts, along
with a record of what type of object they are, and gathering any relevant metadata.

Content is not, of itself, concerned with what you _do_ with this decompositon.
That is handled via the "packer" protocol (for us, the relevant implementation of this
is "bags").
"""

from __future__ import annotations

import abc
import collections.abc
import dataclasses
import operator
import types
from collections.abc import Callable, Iterable, Iterator, Sized
from typing import (
    Any,
    ClassVar,
    Generic,
    Protocol,
    Self,
    SupportsIndex,
    TypeAlias,
    TypeVar,
    cast,
)

import bidict
import h5py
from pyiron_snippets import retrieve

from bagofholding.exceptions import (
    ModuleForbiddenError,
    NoVersionError,
    PickleProtocolError,
    StringNotImportableError,
)
from bagofholding.metadata import (
    Metadata,
    VersionScrapingMap,
    VersionValidatorType,
    get_module,
    get_qualname,
    get_version,
    validate_version,
)

PackingMemoAlias: TypeAlias = bidict.bidict[int, str]
ReferencesAlias: TypeAlias = list[object]
UnpackingMemoAlias: TypeAlias = dict[str, Any]


PackingType = TypeVar("PackingType", bound=Any)
UnpackingType = TypeVar("UnpackingType", bound=Any)


MAX_PICKLE_PROTOCOL = 4
# Although many of the same patterns as `pickle` are exploited to decompose data,
# `bagofholding` does not actually _execute_ `pickle`.
# To this end, the highest protocol value exploiting out-of-band data is not supported


class HasContents(Sized, Iterable[str], Protocol): ...


class Packer(Protocol):
    def join(self, *paths: str) -> str: ...

    def pack_metadata(self, metadata: Metadata, path: str) -> None: ...
    def unpack_metadata(self, path: str) -> Metadata: ...

    def pack_empty(self, path: str) -> None: ...

    def pack_string(self, obj: str, path: str) -> None: ...
    def unpack_string(self, path: str) -> str: ...

    def pack_bool(self, obj: bool, path: str) -> None: ...
    def unpack_bool(self, path: str) -> bool: ...

    def pack_long(self, obj: int, path: str) -> None: ...
    def unpack_long(self, path: str) -> int: ...

    def pack_float(self, obj: float, path: str) -> None: ...
    def unpack_float(self, path: str) -> float: ...

    def pack_complex(self, obj: complex, path: str) -> None: ...
    def unpack_complex(self, path: str) -> complex: ...

    def pack_bytes(self, obj: bytes, path: str) -> None: ...
    def unpack_bytes(self, path: str) -> bytes: ...

    def pack_bytearray(self, obj: bytearray, path: str) -> None: ...
    def unpack_bytearray(self, path: str) -> bytearray: ...

    def create_group(self, path: str) -> None: ...
    def open_group(self, path: str) -> HasContents: ...

    def get_bespoke_content_class(
        self, obj: object
    ) -> type[BespokeItem[Any, Self]] | None: ...


@dataclasses.dataclass
class PackingArguments:
    memo: PackingMemoAlias
    references: ReferencesAlias
    require_versions: bool
    forbidden_modules: list[str] | tuple[str, ...]
    version_scraping: VersionScrapingMap | None
    _pickle_protocol: SupportsIndex


@dataclasses.dataclass
class UnpackingArguments:
    memo: UnpackingMemoAlias
    version_validator: VersionValidatorType
    version_scraping: VersionScrapingMap | None


class NotData:
    pass


PackerType = TypeVar("PackerType", bound=Packer)


class Content(Generic[PackingType, UnpackingType, PackerType], abc.ABC):

    _rich_metadata: ClassVar[bool] = False

    @classmethod
    @abc.abstractmethod
    def pack(
        cls,
        obj: PackingType,
        packer: PackerType,
        path: str,
        packing: PackingArguments,
    ) -> None: ...

    @classmethod
    @abc.abstractmethod
    def unpack(
        cls, packer: PackerType, path: str, unpacking: UnpackingArguments
    ) -> UnpackingType: ...

    @classmethod
    def _get_metadata(cls, obj: PackingType, packing: PackingArguments) -> Metadata:
        if cls._rich_metadata:
            module = get_module(obj)
            if module == "builtins":
                return Metadata(cls.full_name())
            else:
                if module.split(".")[0] in packing.forbidden_modules:
                    raise ModuleForbiddenError(
                        f"Module '{module}' is forbidden as a source of stored objects. Change "
                        f"the `forbidden_modules` or move this object to an allowed module."
                    )

                version = get_version(module, packing.version_scraping)
                if packing.require_versions and version is None:
                    raise NoVersionError(
                        f"Could not find a version for {module}. Either disable "
                        f"`require_versions`, use `version_scraping` to find an existing "
                        f"version for this package, or add versioning to the unversioned "
                        f"package."
                    )

                return Metadata(
                    cls.full_name(),
                    qualname=get_qualname(obj),
                    module=module,
                    version=version,
                    meta=(
                        str(obj.__metadata__) if hasattr(obj, "__metadata__") else None
                    ),
                )
        else:
            return Metadata(cls.full_name())

    @classmethod
    def full_name(cls) -> str:
        return cls.__module__ + "." + cls.__name__


class Item(
    Content[PackingType, UnpackingType, PackerType],
    Generic[PackingType, UnpackingType, PackerType],
    abc.ABC,
):
    @classmethod
    def pack(
        cls,
        obj: PackingType,
        packer: PackerType,
        path: str,
        packing: PackingArguments,
    ) -> None:
        cls._pack_item(obj, packer, path)
        packer.pack_metadata(cls._get_metadata(obj, packing), path)

    @classmethod
    @abc.abstractmethod
    def _pack_item(cls, obj: PackingType, packer: PackerType, path: str) -> None: ...


class Reference(Item[str, Any, Packer]):
    @classmethod
    def _pack_item(cls, obj: str, packer: Packer, path: str) -> None:
        packer.pack_string(obj, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> Any:
        reference = packer.unpack_string(path)
        from_memo = unpacking.memo.get(reference, NotData)
        if from_memo is not NotData:
            return from_memo
        else:
            return unpack(
                packer,
                reference,
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            )


GlobalType: TypeAlias = type[type] | types.FunctionType | str


class Global(Item[GlobalType, Any, Packer]):
    _rich_metadata = True

    @classmethod
    def _pack_item(cls, obj: GlobalType, packer: Packer, path: str) -> None:
        value: str
        if isinstance(obj, str):
            value = "builtins." + obj if "." not in obj else obj
        else:
            value = obj.__module__ + "." + obj.__qualname__

        if "<lambda>" in value:
            raise StringNotImportableError(
                f"Lambda functions are not re-importable, can't pack {obj}"
            )
        elif "<locals>" in value:
            raise StringNotImportableError(
                f"Local functions are not re-importable, can't pack {obj}"
            )
        packer.pack_string(value, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> Any:
        import_string = packer.unpack_string(path)
        return retrieve.import_from_string(import_string)


class NoneItem(Item[type[None], None, Packer]):
    @classmethod
    def _pack_item(cls, obj: type[None], packer: Packer, path: str) -> None:
        packer.pack_empty(path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> None:
        return None


ItemType = TypeVar("ItemType", bound=Any)


class ReflexiveItem(
    Item[ItemType, ItemType, PackerType], Generic[ItemType, PackerType], abc.ABC
): ...


BuiltinItemType = TypeVar(
    "BuiltinItemType",
    str,
    bytes,
    bytearray,
    bool,
    int,
    float,
    complex,
)


class BuiltinItem(
    ReflexiveItem[BuiltinItemType, Packer], Generic[BuiltinItemType], abc.ABC
): ...


class Str(BuiltinItem[str]):
    @classmethod
    def _pack_item(cls, obj: str, packer: Packer, path: str) -> None:
        packer.pack_string(obj, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> str:
        return packer.unpack_string(path)


class Bool(BuiltinItem[bool]):
    @classmethod
    def _pack_item(cls, obj: bool, packer: Packer, path: str) -> None:
        packer.pack_bool(obj, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> bool:
        return packer.unpack_bool(path)


class Long(BuiltinItem[int]):
    @classmethod
    def _pack_item(cls, obj: int, packer: Packer, path: str) -> None:
        packer.pack_long(obj, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> int:
        return packer.unpack_long(path)


class Float(BuiltinItem[float]):
    @classmethod
    def _pack_item(cls, obj: float, packer: Packer, path: str) -> None:
        packer.pack_float(obj, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> float:
        return packer.unpack_float(path)


class Complex(BuiltinItem[complex]):
    @classmethod
    def _pack_item(cls, obj: complex, packer: Packer, path: str) -> None:
        packer.pack_complex(obj, path)

    @classmethod
    def unpack(
        cls, packer: Packer, path: str, unpacking: UnpackingArguments
    ) -> complex:
        return packer.unpack_complex(path)


class Bytes(BuiltinItem[bytes]):
    @classmethod
    def _pack_item(cls, obj: bytes, packer: Packer, path: str) -> None:
        packer.pack_bytes(obj, path)

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> bytes:
        return packer.unpack_bytes(path)


class Bytearray(BuiltinItem[bytearray]):
    @classmethod
    def _pack_item(cls, obj: bytearray, packer: Packer, path: str) -> None:
        packer.pack_bytearray(obj, path)

    @classmethod
    def unpack(
        cls, packer: Packer, path: str, unpacking: UnpackingArguments
    ) -> bytearray:
        return packer.unpack_bytearray(path)


class BespokeItem(
    ReflexiveItem[ItemType, PackerType], Generic[ItemType, PackerType], abc.ABC
):
    _rich_metadata = True


class Group(
    Content[PackingType, UnpackingType, Packer],
    Generic[PackingType, UnpackingType],
    abc.ABC,
): ...


GroupType = TypeVar("GroupType", bound=Any)  # Bind to container?


class ReflexiveGroup(Group[GroupType, GroupType], Generic[GroupType], abc.ABC): ...


# __reduce__ return values
# per https://docs.python.org/3/library/pickle.html#object.__reduce__
ConstructorType: TypeAlias = Callable[..., object]
ConstructorArgsType: TypeAlias = tuple[object, ...]
StateType: TypeAlias = object
ListItemsType: TypeAlias = Iterator[object]
DictItemsType: TypeAlias = Iterator[tuple[object, object]]
SetStateCallableType: TypeAlias = Callable[[object, object], None]
ReduceReturnType: TypeAlias = (
    tuple[ConstructorType, ConstructorArgsType]
    | tuple[ConstructorType, ConstructorArgsType, StateType | None]
    | tuple[
        ConstructorType, ConstructorArgsType, StateType | None, ListItemsType | None
    ]
    | tuple[
        ConstructorType,
        ConstructorArgsType,
        StateType | None,
        ListItemsType | None,
        DictItemsType | None,
    ]
    | tuple[
        ConstructorType | None,
        ConstructorArgsType | None,
        StateType | None,
        ListItemsType | None,
        DictItemsType | None,
        SetStateCallableType | None,
    ]
)
PickleHint: TypeAlias = str | tuple[Any, ...]


class Reducible(ReflexiveGroup[object]):
    _rich_metadata = True
    reduction_fields: ClassVar[tuple[str, str, str, str, str, str]] = (
        "constructor",
        "args",
        "state",
        "item_iterator",
        "kv_iterator",
        "setter",
    )

    @classmethod
    def pack(
        cls,
        obj: object,
        packer: Packer,
        path: str,
        packing: PackingArguments,
        rv: ReduceReturnType | None = None,
    ) -> None:
        reduced_value = (
            obj.__reduce_ex__(packing._pickle_protocol) if rv is None else rv
        )
        packer.create_group(path)
        packer.pack_metadata(cls._get_metadata(obj, packing), path)
        for subpath, value in zip(cls.reduction_fields, reduced_value, strict=False):
            pack(
                value,
                packer,
                packer.join(path, subpath),
                packing.memo,
                packing.references,
                packing.require_versions,
                packing.forbidden_modules,
                packing.version_scraping,
                _pickle_protocol=packing._pickle_protocol,
            )

    @classmethod
    def unpack(cls, packer: Packer, path: str, unpacking: UnpackingArguments) -> object:
        constructor = cast(
            ConstructorType,
            unpack(
                packer,
                packer.join(path, "constructor"),
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            ),
        )
        constructor_args = cast(
            ConstructorArgsType,
            unpack(
                packer,
                packer.join(path, "args"),
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            ),
        )
        obj: object = constructor(*constructor_args)
        unpacking.memo[path] = obj
        rv = (constructor, constructor_args) + tuple(
            unpack(
                packer,
                packer.join(path, k),
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            )
            for k in cls.reduction_fields[2 : len(packer.open_group(path))]
        )
        n_items = len(rv)
        if n_items >= 3 and rv[2] is not None:
            if n_items == 6 and rv[5] is not None:
                cast(SetStateCallableType, rv[5])(obj, rv[2])
            elif hasattr(obj, "__setstate__"):
                obj.__setstate__(rv[2])
            else:
                # "If the object has no such method then, the value must be a dictionary"
                obj.__dict__.update(cast(dict[Any, Any], rv[2]))
        if n_items >= 4 and rv[3] is not None:
            if hasattr(obj, "append"):
                for item in cast(ListItemsType, rv[3]):
                    obj.append(item)
            elif hasattr(obj, "extend"):
                obj.extend(list(cast(ListItemsType, rv[3])))
                # TODO: look into efficiency choices for optional usage of extend even
                #  when append exists
            else:
                raise AttributeError(f"{obj} has neither append nor extend methods")
        if n_items >= 5 and rv[4] is not None and hasattr(obj, "__setitem__"):
            for k, v in cast(DictItemsType, rv[4]):
                obj[k] = v

        return obj


BuiltinGroupType = TypeVar(
    "BuiltinGroupType",
    dict[Any, Any],
    dict[str, Any],
    types.UnionType,
    tuple[Any, ...],
    list[Any],
    set[Any],
    frozenset[Any],
)


class BuiltinGroup(
    ReflexiveGroup[BuiltinGroupType], Generic[BuiltinGroupType], abc.ABC
):
    @classmethod
    def pack(
        cls,
        obj: PackingType,
        packer: Packer,
        path: str,
        packing: PackingArguments,
    ) -> None:
        packer.create_group(path)
        packer.pack_metadata(cls._get_metadata(obj, packing), path)
        cls._pack_subcontent(obj, packer, path, packing)

    @classmethod
    @abc.abstractmethod
    def _pack_subcontent(
        cls,
        obj: PackingType,
        packer: Packer,
        path: str,
        packing: PackingArguments,
    ) -> h5py.Group: ...


class Dict(BuiltinGroup[dict[Any, Any]]):
    @classmethod
    def _pack_subcontent(
        cls,
        obj: dict[Any, Any],
        packer: Packer,
        path: str,
        packing: PackingArguments,
    ) -> None:
        pack(
            tuple(obj.keys()),
            packer,
            packer.join(path, "keys"),
            packing.memo,
            packing.references,
            packing.require_versions,
            packing.forbidden_modules,
            packing.version_scraping,
            _pickle_protocol=packing._pickle_protocol,
        )
        pack(
            tuple(obj.values()),
            packer,
            packer.join(path, "values"),
            packing.memo,
            packing.references,
            packing.require_versions,
            packing.forbidden_modules,
            packing.version_scraping,
            _pickle_protocol=packing._pickle_protocol,
        )

    @classmethod
    def unpack(
        cls, packer: Packer, path: str, unpacking: UnpackingArguments
    ) -> dict[Any, Any]:
        return dict(
            zip(
                cast(
                    tuple[Any],
                    unpack(
                        packer,
                        packer.join(path, "keys"),
                        unpacking.memo,
                        version_validator=unpacking.version_validator,
                        version_scraping=unpacking.version_scraping,
                    ),
                ),
                cast(
                    tuple[Any],
                    unpack(
                        packer,
                        packer.join(path, "values"),
                        unpacking.memo,
                        version_validator=unpacking.version_validator,
                        version_scraping=unpacking.version_scraping,
                    ),
                ),
                strict=True,
            )
        )


class StrKeyDict(BuiltinGroup[dict[str, Any]]):
    @classmethod
    def _pack_subcontent(
        cls,
        obj: dict[str, Any],
        packer: Packer,
        path: str,
        packing: PackingArguments,
    ) -> None:
        for k, v in obj.items():
            pack(
                v,
                packer,
                packer.join(path, k),
                packing.memo,
                packing.references,
                packing.require_versions,
                packing.forbidden_modules,
                packing.version_scraping,
                _pickle_protocol=packing._pickle_protocol,
            )

    @classmethod
    def unpack(
        cls, packer: Packer, path: str, unpacking: UnpackingArguments
    ) -> dict[str, Any]:
        return {
            k: unpack(
                packer,
                packer.join(path, k),
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            )
            for k in packer.open_group(path)
        }


class Union(BuiltinGroup[types.UnionType]):
    """
    :class:`types.UnionType` has no :meth:`__reduce__` method. Pickle actually gets
    around this with bespoke logic, and so we need to too.
    """

    @classmethod
    def _pack_subcontent(
        cls,
        obj: types.UnionType,
        packer: Packer,
        path: str,
        packing: PackingArguments,
    ) -> None:
        for i, v in enumerate(obj.__args__):
            pack(
                v,
                packer,
                packer.join(path, f"i{i}"),
                packing.memo,
                packing.references,
                packing.require_versions,
                packing.forbidden_modules,
                packing.version_scraping,
                _pickle_protocol=packing._pickle_protocol,
            )

    @staticmethod
    def _recursive_or(args: collections.abc.Iterable[object]) -> types.UnionType:
        it = iter(args)
        try:
            first = next(it)
            second = next(it)
        except StopIteration:
            raise ValueError("Expected at least two elements for a UnionType") from None

        union: types.UnionType = operator.or_(first, second)

        for arg in it:
            union = operator.or_(union, arg)

        return union

    @classmethod
    def unpack(
        cls, packer: Packer, path: str, unpacking: UnpackingArguments
    ) -> types.UnionType:
        return cls._recursive_or(
            unpack(
                packer,
                packer.join(path, f"i{i}"),
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            )
            for i in range(len(packer.open_group(path)))
        )


IndexableType = TypeVar(
    "IndexableType", tuple[Any, ...], list[Any], set[Any], frozenset[Any]
)


class Indexable(BuiltinGroup[IndexableType], Generic[IndexableType], abc.ABC):
    recast: type[IndexableType]

    @classmethod
    def _pack_subcontent(
        cls,
        obj: IndexableType,
        packer: Packer,
        path: str,
        packing: PackingArguments,
    ) -> None:
        for i, v in enumerate(obj):
            pack(
                v,
                packer,
                packer.join(path, f"i{i}"),
                packing.memo,
                packing.references,
                packing.require_versions,
                packing.forbidden_modules,
                packing.version_scraping,
                _pickle_protocol=packing._pickle_protocol,
            )

    @classmethod
    def unpack(
        cls, packer: Packer, path: str, unpacking: UnpackingArguments
    ) -> IndexableType:
        return cls.recast(
            unpack(
                packer,
                packer.join(path, f"i{i}"),
                unpacking.memo,
                version_validator=unpacking.version_validator,
                version_scraping=unpacking.version_scraping,
            )
            for i in range(len(packer.open_group(path)))
        )


class Tuple(Indexable[tuple[Any, ...]]):
    recast = tuple


class List(Indexable[list[Any]]):
    recast = list


class Set(Indexable[set[Any]]):
    recast = set


class FrozenSet(Indexable[frozenset[Any]]):
    recast = frozenset


def pack(
    obj: object,
    packer: Packer,
    path: str,
    memo: PackingMemoAlias,
    references: ReferencesAlias,
    require_versions: bool,
    forbidden_modules: list[str] | tuple[str, ...],
    version_scraping: VersionScrapingMap | None,
    _pickle_protocol: SupportsIndex = MAX_PICKLE_PROTOCOL,
) -> None:
    if _pickle_protocol not in (4, 3, 2, 1, 0):
        raise PickleProtocolError(
            f"pickle protocol must be <= 4, got {_pickle_protocol}"
        )

    packing_args = PackingArguments(
        memo=memo,
        references=references,
        require_versions=require_versions,
        forbidden_modules=forbidden_modules,
        version_scraping=version_scraping,
        _pickle_protocol=_pickle_protocol,
    )

    t = type if isinstance(obj, type) else type(obj)
    simple_class = KNOWN_ITEM_MAP.get(t)
    if simple_class is not None:
        simple_class.pack(
            obj,
            packer,
            path,
            packing_args,
        )
        return

    obj_id = id(obj)
    reference = memo.get(obj_id)
    if reference is not None:
        Reference.pack(reference, packer, path, packing_args)
        return
    else:
        memo[obj_id] = path
        references.append(obj)

    complex_class = packer.get_bespoke_content_class(obj)
    if complex_class is not None:
        complex_class.pack(obj, packer, path, packing_args)
        return

    group_class = get_group_content_class(obj)
    if group_class is not None:
        group_class.pack(obj, packer, path, packing_args)
        return

    rv = obj.__reduce_ex__(_pickle_protocol)
    if isinstance(rv, str):
        Global.pack(
            retrieve.get_importable_string_from_string_reduction(rv, obj),
            packer,
            path,
            packing_args,
        )
        return
    else:
        Reducible.pack(obj, packer, path, packing_args, rv=rv)
        return


KNOWN_ITEM_MAP: dict[
    type | types.FunctionType | types.BuiltinFunctionType, type[Item[Any, Any, Packer]]
] = {
    type: Global,
    types.FunctionType: Global,
    type(all): Global,
    type(None): NoneItem,
    bool: Bool,
    int: Long,
    float: Float,
    complex: Complex,
    bytes: Bytes,
    bytearray: Bytearray,
    str: Str,
}


KNOWN_GROUP_MAP: dict[type, type[Group[Any, Any]]] = {
    dict: Dict,
    types.UnionType: Union,
    tuple: Tuple,
    list: List,
    set: Set,
    frozenset: FrozenSet,
}


def is_simple_string(string: str) -> bool:
    return string.isascii() and ("/" not in string)


def has_surrogates(s: str) -> bool:
    try:
        s.encode("utf-8")
        return False
    except UnicodeEncodeError:
        return True


def get_group_content_class(obj: object) -> type[Group[Any, Any]] | None:
    t = type(obj)
    if t is dict and all(
        isinstance(k, str) and is_simple_string(k) for k in cast(dict[str, Any], obj)
    ):
        return StrKeyDict

    return KNOWN_GROUP_MAP.get(t)


def unpack(
    packer: Packer,
    path: str,
    memo: UnpackingMemoAlias,
    version_validator: VersionValidatorType,
    version_scraping: VersionScrapingMap | None,
) -> object:
    memo_value = memo.get(path, NotData)
    if memo_value is NotData:
        metadata = packer.unpack_metadata(path)
        content_class = retrieve.import_from_string(metadata.content_type)
        if metadata is not None:
            validate_version(
                metadata, validator=version_validator, version_scraping=version_scraping
            )
        value = content_class.unpack(
            packer,
            path,
            UnpackingArguments(
                memo=memo,
                version_validator=version_validator,
                version_scraping=version_scraping,
            ),
        )
        if path not in memo:
            memo[path] = value
        return value
    return memo_value



"""
h5py supports efficient treatment of a subset of possible array.

Here, we whitelist these types.
"""

from typing import TypeAlias

import numpy as np

H5PY_DTYPE_WHITELIST = (
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.float16,
    np.float32,
    np.float64,
    np.complex64,
    np.complex128,
    np.bool_,
    # np.bytes_,
    # np.str_,
    # h5py needs special handling for byte and string info
)


H5Scalar: TypeAlias = (
    np.int8
    | np.int16
    | np.int32
    | np.int64
    | np.uint8
    | np.uint16
    | np.uint32
    | np.uint64
    | np.float16
    | np.float32
    | np.float64
    | np.complex64
    | np.complex128
    | np.bool_
    | np.bytes_
    | np.str_
)


IntTypesAlias: TypeAlias = (
    np.dtype[np.int8]
    | np.dtype[np.int16]
    | np.dtype[np.int32]
    | np.dtype[np.int64]
    | np.dtype[np.uint8]
    | np.dtype[np.uint16]
    | np.dtype[np.uint32]
    | np.dtype[np.uint64]
)


H5DtypeAlias: TypeAlias = (
    IntTypesAlias
    | np.dtype[np.float16]
    | np.dtype[np.float32]
    | np.dtype[np.float64]
    | np.dtype[np.complex64]
    | np.dtype[np.complex128]
    | np.dtype[np.bool_]
    | np.dtype[np.bytes_]
    | np.dtype[np.str_]
)



"""
The core user-facing object.

Full implementations of bags should guarantee the key features promised by the package:
- Storage and retrieval of arbitrary pickleable python objects
- Metadata preservation
- Versioning verification
- Browsing without loading
- Partial reloading
"""

from __future__ import annotations

import abc
import dataclasses
import os.path
import pathlib
import pickle
from collections.abc import Iterator, Mapping
from typing import (
    Any,
    ClassVar,
    Self,
    SupportsIndex,
)

import bidict
from pyiron_snippets import import_alarm

from bagofholding.content import MAX_PICKLE_PROTOCOL, BespokeItem, Packer, pack, unpack
from bagofholding.exceptions import BagMismatchError, InvalidMetadataError
from bagofholding.metadata import (
    HasFieldIterator,
    HasVersionInfo,
    Metadata,
    VersionScrapingMap,
    VersionValidatorType,
    get_version,
)

try:
    from bagofholding.widget import BagTree

    alarm = import_alarm.ImportAlarm()
except (ImportError, ModuleNotFoundError):
    alarm = import_alarm.ImportAlarm(
        "The browsing widget relies on ipytree and traitlets, but this was "
        "unavailable. You can get a text-representation of all available paths with "
        ":meth:`bagofholding.bag.Bag.list_paths`.",
        raise_exception=True,
    )

PATH_DELIMITER = "/"


@dataclasses.dataclass(frozen=True)
class BagInfo(HasVersionInfo, HasFieldIterator):
    pass


class Bag(Packer, Mapping[str, Metadata | None], abc.ABC):
    """
    Bags are the user-facing object.
    """

    bag_info: BagInfo
    storage_root: ClassVar[str] = "object"
    filepath: pathlib.Path

    @classmethod
    def get_bag_info(cls) -> BagInfo:
        return BagInfo(
            qualname=cls.__qualname__,
            module=cls.__module__,
            version=cls.get_version(),
        )

    @classmethod
    def _bag_info_class(cls) -> type[BagInfo]:
        return BagInfo

    @classmethod
    def save(
        cls,
        obj: Any,
        filepath: str | pathlib.Path,
        require_versions: bool = False,
        forbidden_modules: list[str] | tuple[str, ...] = (),
        version_scraping: VersionScrapingMap | None = None,
        _pickle_protocol: SupportsIndex = MAX_PICKLE_PROTOCOL,
        overwrite_existing: bool = True,
    ) -> None:
        """
        Save a python object to file.

        Args:
            obj (Any): The (pickleble) python object to be saved.
            filepath (str|pathlib.Path): The path to save the object to.
            require_versions (bool): Whether to require a metadata for reduced
                and complex objects to contain a non-None version. (Default is False,
                objects can be stored from non-versioned packages/modules.)
            forbidden_modules (list[str] | tuple[str, ...] | None): Do not allow saving
                objects whose root-most modules are listed here. (Default is an empty
                tuple, i.e. don't disallow anything.) This is particularly useful to
                disallow  `"__main__"` to improve the odds that objects will actually
                be loadable in the future.
            version_scraping (dict[str, Callable[[str], str]] | None): An optional
                dictionary mapping module names to a callable that takes this name and
                returns a version (or None). The default callable imports the module
                string and looks for a `__version__` attribute.
        """
        if os.path.exists(filepath):
            if overwrite_existing and os.path.isfile(filepath):
                os.remove(filepath)
            else:
                raise FileExistsError(f"{filepath} already exists or is not a file.")
        bag = cls(filepath)
        bag._pack_bag_info()
        pack(
            obj,
            bag,
            bag.storage_root,
            bidict.bidict(),
            [],
            require_versions,
            forbidden_modules,
            version_scraping,
            _pickle_protocol=_pickle_protocol,
        )
        bag._write()

    @classmethod
    def get_version(cls) -> str:
        return str(get_version(cls.__module__, {}))

    def __init__(
        self, filepath: str | pathlib.Path, *args: object, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.filepath = pathlib.Path(filepath)
        if os.path.isfile(self.filepath):
            self.bag_info = self._unpack_bag_info()
            if not self.validate_bag_info(self.bag_info, self.get_bag_info()):
                raise BagMismatchError(
                    f"The bag class {self.__class__} does not match the bag saved at "
                    f"{filepath}; class info is {self.get_bag_info()}, but the info saved "
                    f"is {self.bag_info}"
                )

    @abc.abstractmethod
    def _pack_field(self, path: str, key: str, value: str) -> None: ...

    @abc.abstractmethod
    def _unpack_field(self, path: str, key: str) -> str | None: ...

    @staticmethod
    def validate_bag_info(bag_info: BagInfo, reference: BagInfo) -> bool:
        return bag_info == reference

    def load(
        self,
        path: str = storage_root,
        version_validator: VersionValidatorType = "exact",
        version_scraping: VersionScrapingMap | None = None,
    ) -> Any:
        return unpack(
            self,
            path,
            {},
            version_validator=version_validator,
            version_scraping=version_scraping,
        )

    def __getitem__(self, path: str) -> Metadata:
        return self.unpack_metadata(path)

    @abc.abstractmethod
    def list_paths(self) -> list[str]:
        """A list of all available content paths."""

    @alarm
    def widget(self):  # type: ignore[no-untyped-def]
        return BagTree(self)

    def browse(self):  # type: ignore[no-untyped-def]
        try:
            return self.widget()
        except ImportError:
            return self.list_paths()

    def __len__(self) -> int:
        return len(self.list_paths())

    def __iter__(self) -> Iterator[str]:
        return iter(self.list_paths())

    def join(self, *paths: str) -> str:
        return PATH_DELIMITER.join(paths)

    @staticmethod
    def pickle_check(
        obj: Any, raise_exceptions: bool = True, print_message: bool = False
    ) -> str | None:
        """
        A simple helper to check if an object can be pickled and unpickled.
        Useful if you run into trouble saving or loading and want to see whether the
        underlying object is compliant with pickle-ability requirements to begin with.

        Args:
            obj: The object to test for pickling support.
            raise_exceptions: If True, re-raise any exception encountered.
            print_message: If True, print the exception message on failure.

        Returns:
            None if pickling is successful; otherwise, returns the exception message as a string.
        """

        try:
            pickle.loads(pickle.dumps(obj))
        except Exception as e:
            if print_message:
                print(e)
            if raise_exceptions:
                raise e
            return str(e)
        return None

    def _pack_fields(self, dataclass: HasFieldIterator, path: str) -> None:
        for k, v in dataclass.field_items():
            if v is not None:
                self._pack_field(path, k, v)

    def _unpack_fields(
        self, dataclass_type: type[HasFieldIterator], path: str
    ) -> dict[str, str | None]:
        field_values: dict[str, str | None] = {}
        for k in dataclass_type.__dataclass_fields__:
            field_values[k] = self._unpack_field(path, k)
        return field_values

    def _pack_bag_info(self) -> None:
        self._pack_fields(self.get_bag_info(), PATH_DELIMITER)

    def _unpack_bag_info(self) -> BagInfo:
        return self._bag_info_class()(
            **self._unpack_fields(self._bag_info_class(), PATH_DELIMITER)
        )

    def _write(self) -> None:
        return

    def pack_metadata(self, metadata: Metadata, path: str) -> None:
        self._pack_fields(metadata, path)
        return None

    def unpack_metadata(self, path: str) -> Metadata:
        metadata = self._unpack_fields(Metadata, path)
        content_type = metadata.pop("content_type", None)
        if content_type is None:
            raise InvalidMetadataError(f"Metadata at {path} is missing a content type")
        return Metadata(content_type, **metadata)

    def get_bespoke_content_class(
        self, obj: object
    ) -> type[BespokeItem[Any, Self]] | None:
        return None



import importlib.metadata

try:
    # Installed package will find its version
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    # Repository clones will register an unknown version
    __version__ = "0.0.0+unknown"

from bagofholding.exceptions import BagMismatchError as BagMismatchError
from bagofholding.exceptions import BagOfHoldingError as BagOfHoldingError
from bagofholding.exceptions import EnvironmentMismatchError as EnvironmentMismatchError
from bagofholding.exceptions import FileAlreadyOpenError as FileAlreadyOpenError
from bagofholding.exceptions import FileNotOpenError as FileNotOpenError
from bagofholding.exceptions import FilepathError as FilepathError
from bagofholding.exceptions import InvalidMetadataError as InvalidMetadataError
from bagofholding.exceptions import ModuleForbiddenError as ModuleForbiddenError
from bagofholding.exceptions import NotAGroupError as NotAGroupError
from bagofholding.exceptions import NoVersionError as NoVersionError
from bagofholding.exceptions import PickleProtocolError as PickleProtocolError
from bagofholding.exceptions import (
    StringNotImportableError as StringNotImportableError,
)
from bagofholding.h5.bag import H5Bag as H5Bag
from bagofholding.h5.triebag import TrieH5Bag as TrieH5Bag



from __future__ import annotations

import pathlib
from typing import Any, ClassVar, Self, TypeAlias, TypeVar, cast

import bidict
import h5py
import numpy as np
import pygtrie

from bagofholding.bag import PATH_DELIMITER, Bag, BagInfo
from bagofholding.content import BespokeItem, has_surrogates
from bagofholding.h5.bag import H5Info
from bagofholding.h5.content import Array, ArrayPacker, ArrayType, int_overflows
from bagofholding.h5.context import HasH5FileContext
from bagofholding.h5.dtypes import H5PY_DTYPE_WHITELIST, H5Scalar, IntTypesAlias
from bagofholding.metadata import Metadata, VersionScrapingMap, VersionValidatorType
from bagofholding.trie import decompose_stringtrie, reconstruct_stringtrie

PackedThingType = TypeVar("PackedThingType", str, bool, int, float, bytes, bytearray)

StringArrayType: TypeAlias = np.ndarray[tuple[int, ...], np.dtype[np.str_]]
IntArrayType: TypeAlias = np.ndarray[tuple[int, ...], IntTypesAlias]


class TrieH5Bag(Bag, HasH5FileContext, ArrayPacker):
    """
    A bag using HDF5 files based on `h5py`.

    Uses a trie structure to flatten the stored object. Compared to
    :class:`bagofholding.h5.bag.H5Bag`, this is advantageous for file sizes but
    (currently)  has worse scaling for save times.

    The resulting HDF5 file cannot be directly related to the structure of the stored
    object, but must be re-mapped via mapping fields. Metadata is also lumped in with
    other string data to minimize the number of different h5 groups.
    """

    _content_key: ClassVar[str] = "content_type"

    _paths_key: ClassVar[str] = "paths"
    _type_index_key: ClassVar[str] = "type_index"
    _position_index_key: ClassVar[str] = "position_index"
    _index_map: ClassVar[bidict.bidict[str, int]] = bidict.bidict(
        {
            # Real codes mapping to pack lists in self._packed
            "str": 0,
            "bool": 1,
            "long": 2,
            "float": 3,
            "complex_real": 4,
            "complex_imag": 5,
            "bytes": 6,
            "bytearray": 7,
            "array": 8,
            "empty": 9,
            "bigint": 10,
            # Synthetic codes mapping to concepts
            "group": 100,
            "empty_bytes": 101,
            "surrogate_str": 102,
        }
    )
    _field_delimiter: ClassVar[str] = "::"
    _child_delimiter: ClassVar[str] = ";"

    @classmethod
    def get_bag_info(cls) -> BagInfo:
        return H5Info(
            qualname=cls.__qualname__,
            module=cls.__module__,
            version=cls.get_version(),
            libver_str=cls.libver_str,
        )

    @classmethod
    def _bag_info_class(cls) -> type[BagInfo]:
        return H5Info

    def __init__(
        self, filepath: str | pathlib.Path, *args: object, **kwargs: Any
    ) -> None:
        self._file = None
        self._context_depth = 0
        self._unpacked_paths: StringArrayType | None = None
        self._unpacked_type_index: IntArrayType | None = None
        self._unpacked_position_index: IntArrayType | None = None
        self._unpacked_nonmetadata_paths: list[str] | None = None
        self._path_to_index: dict[str, int] | None = None
        self._unpacked_trie: pygtrie.StringTrie | None = None
        super().__init__(filepath)
        self._packed_trie: pygtrie.StringTrie = pygtrie.StringTrie()
        self._packed: tuple[
            list[str],
            list[bool],
            list[int],
            list[float],
            list[float],
            list[float],
            list[bytes],
            list[bytearray],
            list[ArrayType],
            list[bytes],
            list[str],
        ] = ([], [], [], [], [], [], [], [], [], [], [])

    @property
    def unpacked_trie(self) -> pygtrie.StringTrie:
        if self._unpacked_trie is None:
            with self:
                self._unpacked_trie = reconstruct_stringtrie(
                    self.file["trie_segments"][:].astype("str"),
                    self.file["trie_parents"][:],
                    [v.tolist() for v in self.file["trie_values"][:]],
                    [-1, -1],
                )
        return self._unpacked_trie

    def _write(self) -> None:
        str_type = h5py.string_dtype(encoding="utf-8")

        self.open("w")
        segments, parents, values = decompose_stringtrie(
            self._packed_trie, null_value=(-1, -1)
        )
        self.file.create_dataset(
            "trie_segments",
            data=np.array(segments, dtype=h5py.string_dtype(encoding="utf-8")),
        )
        self.file.create_dataset("trie_parents", data=np.array(parents, dtype=np.int32))
        self.file.create_dataset("trie_values", data=np.array(values, dtype=np.int32))

        self.file.create_dataset("str", data=np.array(self._packed[0], dtype=str_type))
        self.file.create_dataset("bool", data=np.array(self._packed[1], dtype=bool))
        self.file.create_dataset("long", data=np.array(self._packed[2], dtype=int))
        self.file.create_dataset("float", data=np.array(self._packed[3], dtype=float))
        self.file.create_dataset(
            "complex_real", data=np.array(self._packed[4], dtype=float)
        )
        self.file.create_dataset(
            "complex_imag", data=np.array(self._packed[5], dtype=float)
        )
        bytes_group = self.file.create_group("bytes_data")
        for i, b in enumerate(self._packed[6]):
            bytes_group.create_dataset(f"i{i}", data=np.void(b))
        bytearray_group = self.file.create_group("bytearrays")
        for i, ba in enumerate(self._packed[7]):
            bytearray_group.create_dataset(f"i{i}", data=ba)
        # A reasonable alternative would be to group bytes and bytearrays by their
        # length, creating fewer overall groups. Benefit depends on how heavily weighted
        # unique lengths are to shared lengths, so for now keep it simple
        array_group = self.file.create_group("ndarrays")
        for i, ra in enumerate(self._packed[8]):
            array_group.create_dataset(f"i{i}", data=ra)
        # Empty doesn't need to be packed -- it's always None so the meta info is enough
        # Groups don't need to be packed -- they are just holders so meta info is enough

        # Some string encoding requires special treatment to play with h5py
        surrogate_group = self.file.create_group("surrogate_strs")
        for i, s in enumerate(self._packed[9]):
            surrogate_group.create_dataset(f"i{i}", data=np.void(s))

        # h5py has overflow limits on integers, so store them separately stringified
        self.file.create_dataset(
            "bigint", data=np.array(self._packed[10], dtype=str_type)
        )

        self.close()

    def _unpack_bag_info(self) -> BagInfo:
        with self:
            info = super()._unpack_bag_info()
        return info

    def load(
        self,
        path: str = Bag.storage_root,
        version_validator: VersionValidatorType = "exact",
        version_scraping: VersionScrapingMap | None = None,
    ) -> Any:
        with self:
            unpacked = super().load(
                path=path,
                version_validator=version_validator,
                version_scraping=version_scraping,
            )
        return unpacked

    def __getitem__(self, path: str) -> Metadata:
        with self:
            return super().__getitem__(self._sanitize_path(path))

    def list_paths(self) -> list[str]:
        """A list of all available content paths."""
        if self._unpacked_nonmetadata_paths is None:
            paths = self.unpacked_trie.keys()
            self._unpacked_nonmetadata_paths = [
                self._sanitize_path(p)
                for p in np.array(paths)[
                    ~np.char.find(paths, self._field_delimiter) >= 0
                ]
            ]
        return self._unpacked_nonmetadata_paths

    def __enter__(self) -> Self:
        self._context_depth += 1
        if self._file is None:
            self.open("r")
        return self

    def _pack_trie(self, path: str, type_index: int, position_index: int) -> None:
        self._packed_trie[PATH_DELIMITER + path] = [type_index, position_index]

    def _read_trie(self, path: str) -> tuple[int, int]:
        return cast(
            tuple[int, int],
            self.unpacked_trie.values(prefix=PATH_DELIMITER + path, shallow=True)[0],
        )

    def _field_to_path(self, path: str, key: str) -> str:
        return self._sanitize_path(path) + self._field_delimiter + key

    def _sanitize_path(self, path: str) -> str:
        return path.rstrip(PATH_DELIMITER).lstrip(PATH_DELIMITER)

    def _pack_field(self, path: str, key: str, value: str) -> None:
        type_index = self._index_map["str"]
        data_list = self._packed[type_index]
        data_list.append(value)  # type: ignore[arg-type]
        self._pack_trie(self._field_to_path(path, key), type_index, len(data_list) - 1)

    def _unpack_field(self, path: str, key: str) -> str | None:
        try:
            return self.maybe_decode(
                cast(str, self._read_pathlike(self._field_to_path(path, key)))
            )
        except KeyError:
            return None

    def _read_pathlike(self, path: str) -> H5Scalar:
        # A real path or one with the field delimiter to find a metadata field
        type_index, position_index = self._read_trie(path)
        group_name = self._index_map.inverse[type_index]
        with self:
            value = cast(H5Scalar, self.file[group_name][position_index])
        return value

    def pack_empty(self, path: str) -> None:
        self._pack_trie(path, self._index_map["empty"], -1)

    def _pack_thing(
        self, obj: PackedThingType | ArrayType, type_name: str, path: str
    ) -> None:
        type_index = self._index_map[type_name]
        group = self._packed[type_index]
        group.append(obj)  # type: ignore[arg-type]
        self._pack_trie(path, type_index, len(group) - 1)

    def pack_string(self, obj: str, path: str) -> None:
        if has_surrogates(obj):
            encoded = obj.encode("utf-16", errors="surrogatepass")
            type_index = self._index_map["surrogate_str"]
            group = self._packed[9]
            group.append(encoded)
            self._pack_trie(path, type_index, len(group) - 1)
        else:
            self._pack_thing(obj, "str", path)

    def unpack_string(self, path: str) -> str:
        type_index, position_index = self._read_trie(path)
        if self._index_map.inverse[type_index] == "surrogate_str":
            with self:
                data = self.file[f"surrogate_strs/i{position_index}"][()]
            return cast(str, data.tobytes().decode("utf-16", errors="surrogatepass"))
        return self.maybe_decode(cast(str, self._read_pathlike(path)))

    def pack_bool(self, obj: bool, path: str) -> None:
        self._pack_thing(obj, "bool", path)

    def unpack_bool(self, path: str) -> bool:
        return bool(self._read_pathlike(path))

    def pack_long(self, obj: int, path: str) -> None:
        if int_overflows(obj):
            self._pack_thing(str(obj), "bigint", path)
        else:
            self._pack_thing(obj, "long", path)

    def unpack_long(self, path: str) -> int:
        return int(self._read_pathlike(path))

    def pack_float(self, obj: float, path: str) -> None:
        self._pack_thing(obj, "float", path)

    def unpack_float(self, path: str) -> float:
        return float(self._read_pathlike(path))

    def pack_complex(self, obj: complex, path: str) -> None:
        real_index = self._index_map["complex_real"]
        real_group = self._packed[real_index]
        real_group.append(obj.real)  # type: ignore[arg-type]
        imag_index = self._index_map["complex_imag"]
        imag_group = self._packed[imag_index]
        imag_group.append(obj.imag)  # type: ignore[arg-type]
        self._pack_trie(path, real_index, len(real_group) - 1)

    def unpack_complex(self, path: str) -> complex:
        _, position_index = self._read_trie(path)
        with self:
            value = complex(
                self.file["complex_real"][position_index],
                self.file["complex_imag"][position_index],
            )
        return value

    def pack_bytes(self, obj: bytes, path: str) -> None:
        if obj == b"":
            self._pack_trie(path, self._index_map["empty_bytes"], -1)
        else:
            self._pack_thing(obj, "bytes", path)

    def unpack_bytes(self, path: str) -> bytes:
        type_index, position_index = self._read_trie(path)
        if self._index_map.inverse[type_index] == "empty_bytes":
            return b""
        else:
            with self:
                return bytes(self.file[f"bytes_data/i{position_index}"][()])

    def pack_bytearray(self, obj: bytearray, path: str) -> None:
        self._pack_thing(obj, "bytearray", path)

    def unpack_bytearray(self, path: str) -> bytearray:
        _, position_index = self._read_trie(path)
        with self:
            value = bytearray(self.file[f"bytearrays/i{position_index}"][()])
        return value

    def create_group(self, path: str) -> None:
        self._pack_trie(path, self._index_map["group"], -1)

    def open_group(self, path: str) -> set[str]:
        prefix = PATH_DELIMITER + path
        subpaths = self.unpacked_trie.keys(prefix=prefix, shallow=False)
        next_depth_index = 1
        children = {
            part[next_depth_index]
            for key in subpaths
            if (part := key[len(prefix) :].split(PATH_DELIMITER, next_depth_index + 1))
            and len(part) > next_depth_index
            and self._field_delimiter not in part[next_depth_index]
        }
        return children

    # def get_bespoke_content_class(self, obj: object) -> type[BespokeItem[Any, Self]] | None:
    def get_bespoke_content_class(
        self, obj: object
    ) -> type[BespokeItem[Any, Self]] | None:
        if type(obj) is np.ndarray and obj.dtype.type in H5PY_DTYPE_WHITELIST:
            return cast(type[BespokeItem[Any, Self]], Array)
        return None

    def pack_array(self, obj: ArrayType, path: str) -> None:
        self._pack_thing(obj, "array", path)

    def unpack_array(self, path: str) -> ArrayType:
        _, position_index = self._read_trie(path)
        with self:
            value = cast(ArrayType, self.file[f"ndarrays/i{position_index}"][:])
        return value



from __future__ import annotations

import dataclasses
import pathlib
from typing import Any, ClassVar, Self, cast

import h5py
import numpy as np

from bagofholding.bag import Bag, BagInfo
from bagofholding.content import BespokeItem, has_surrogates
from bagofholding.exceptions import NotAGroupError
from bagofholding.h5.content import Array, ArrayPacker, ArrayType, int_overflows
from bagofholding.h5.context import HasH5FileContext
from bagofholding.h5.dtypes import H5PY_DTYPE_WHITELIST
from bagofholding.metadata import Metadata, VersionScrapingMap, VersionValidatorType


@dataclasses.dataclass(frozen=True)
class H5Info(BagInfo):
    libver_str: str = "latest"


class H5Bag(Bag, HasH5FileContext, ArrayPacker):
    """
    A bag using HDF5 files based on `h5py`.

    The underlying file structure is directly representative of the structure of the
    decomposed object being stored, and `attrs` are used to store metadata.
    """

    _content_key: ClassVar[str] = "content_type"

    @classmethod
    def get_bag_info(cls) -> BagInfo:
        return H5Info(
            qualname=cls.__qualname__,
            module=cls.__module__,
            version=cls.get_version(),
            libver_str=cls.libver_str,
        )

    @classmethod
    def _bag_info_class(cls) -> type[BagInfo]:
        return H5Info

    def __init__(
        self, filepath: str | pathlib.Path, *args: object, **kwargs: Any
    ) -> None:
        self._file = None
        self._context_depth = 0
        super().__init__(filepath)

    def _write(self) -> None:
        self.close()

    def _pack_bag_info(self) -> None:
        self.open("w")
        super()._pack_bag_info()

    def _unpack_bag_info(self) -> BagInfo:
        with self:
            info = super()._unpack_bag_info()
        return info

    def load(
        self,
        path: str = Bag.storage_root,
        version_validator: VersionValidatorType = "exact",
        version_scraping: VersionScrapingMap | None = None,
    ) -> Any:
        with self:
            unpacked = super().load(
                path=path,
                version_validator=version_validator,
                version_scraping=version_scraping,
            )
        return unpacked

    def __getitem__(self, path: str) -> Metadata:
        with self:
            return super().__getitem__(path)

    def list_paths(self) -> list[str]:
        """A list of all available content paths."""
        paths: list[str] = []
        with self:
            self.file.visit(paths.append)
        return paths

    def __enter__(self) -> Self:
        self._context_depth += 1
        if self._file is None:
            self.open("r")
        return self

    def _pack_field(self, path: str, key: str, value: str) -> None:
        self.file[path].attrs[key] = value

    def _unpack_field(self, path: str, key: str) -> str | None:
        try:
            return self.maybe_decode(self.file[path].attrs[key])
        except KeyError:
            return None

    def pack_empty(self, path: str) -> None:
        self.file.create_dataset(path, data=h5py.Empty(dtype="f"))

    def pack_string(self, obj: str, path: str) -> None:
        if has_surrogates(obj):
            encoded = obj.encode("utf-16", errors="surrogatepass")
            self.file.create_dataset(path, data=np.void(encoded))
            self.file[path].attrs["_surrogate_str"] = True
        else:
            self.file.create_dataset(
                path, data=obj, dtype=h5py.string_dtype(encoding="utf-8")
            )

    def unpack_string(self, path: str) -> str:
        if self.file[path].attrs.get("_surrogate_str", False):
            return cast(
                str,
                self.file[path][()].tobytes().decode("utf-16", errors="surrogatepass"),
            )
        return cast(str, self._unpack_raw(path).decode("utf-8"))

    def _pack_raw(self, obj: bytearray | bool | int | float, path: str) -> None:
        self.file.create_dataset(path, data=obj)

    def _unpack_raw(self, path: str) -> Any:
        return self.file[path][()]

    def pack_bool(self, obj: bool, path: str) -> None:
        return self._pack_raw(obj, path)

    def unpack_bool(self, path: str) -> bool:
        return bool(self._unpack_raw(path))

    def pack_long(self, obj: int, path: str) -> None:
        if int_overflows(obj):
            self.file.create_dataset(
                path, data=str(obj), dtype=h5py.string_dtype(encoding="utf-8")
            )
            self.file[path].attrs["_bigint"] = True
        else:
            self._pack_raw(obj, path)

    def unpack_long(self, path: str) -> int:
        if self.file[path].attrs.get("_bigint", False):
            return int(self._unpack_raw(path).decode("utf-8"))
        return int(self._unpack_raw(path))

    def pack_float(self, obj: float, path: str) -> None:
        return self._pack_raw(obj, path)

    def unpack_float(self, path: str) -> float:
        return float(self._unpack_raw(path))

    def pack_complex(self, obj: complex, path: str) -> None:
        self.file.create_dataset(path, data=np.array([obj.real, obj.imag]))

    def unpack_complex(self, path: str) -> complex:
        data = self._unpack_raw(path)
        return complex(data[0], data[1])

    def pack_bytes(self, obj: bytes, path: str) -> None:
        if obj == b"":
            special = h5py.special_dtype(vlen=bytes)
            self.file.create_dataset(path, data=b"", dtype=special)
        else:
            self.file.create_dataset(path, data=np.void(obj))

    def unpack_bytes(self, path: str) -> bytes:
        return bytes(self._unpack_raw(path))

    def pack_bytearray(self, obj: bytearray, path: str) -> None:
        return self._pack_raw(obj, path)

    def unpack_bytearray(self, path: str) -> bytearray:
        return bytearray(self._unpack_raw(path))

    def create_group(self, path: str) -> None:
        self.file.create_group(path)

    def open_group(self, path: str) -> list[str]:
        with self:
            group = self.file[path]
            if not isinstance(group, h5py.Group):
                raise NotAGroupError(f"Asked a group at {path}, got {type(group)}")
            subcontent_names = list(group)
        return subcontent_names

    # def get_bespoke_content_class(self, obj: object) -> type[BespokeItem[Any, Self]] | None:
    def get_bespoke_content_class(
        self, obj: object
    ) -> type[BespokeItem[Any, Self]] | None:
        if type(obj) is np.ndarray and obj.dtype.type in H5PY_DTYPE_WHITELIST:
            return cast(type[BespokeItem[Any, Self]], Array)
        return None

    def pack_array(self, obj: ArrayType, path: str) -> None:
        self.file.create_dataset(path, data=obj)

    def unpack_array(self, path: str) -> ArrayType:
        return cast(ArrayType, self.file[path][()])



"""
Tools for extracting and logging information about python objects.
"""

from __future__ import annotations

import dataclasses
import re
from collections.abc import Callable, ItemsView
from importlib import import_module
from sys import version_info
from typing import Any, Literal, TypeAlias

from bagofholding.exceptions import EnvironmentMismatchError


@dataclasses.dataclass(frozen=True)
class HasFieldIterator:
    """A simple helper mixin for dataclasses"""

    def field_items(self) -> ItemsView[str, str | None]:
        return dataclasses.asdict(self).items()


@dataclasses.dataclass(frozen=True)
class HasContentType:
    content_type: str


@dataclasses.dataclass(frozen=True)
class HasVersionInfo:
    qualname: str | None = None
    module: str | None = None
    version: str | None = None


@dataclasses.dataclass(frozen=True)
class Metadata(HasVersionInfo, HasContentType, HasFieldIterator):
    meta: str | None = None


def get_module(obj: Any) -> str:
    return obj.__module__ if isinstance(obj, type) else type(obj).__module__


def get_qualname(obj: Any) -> str:
    return obj.__qualname__ if isinstance(obj, type) else type(obj).__qualname__


VersionScraperType: TypeAlias = Callable[[str], str | None]
VersionScrapingMap: TypeAlias = dict[str, VersionScraperType]


def get_version(
    module_name: str,
    version_scraping: VersionScrapingMap | None = None,
) -> str | None:
    """
    Given a module name, get its associated version (if any). By default, this simply
    looks for the :attr:`__version__` attribute on the imported module.

    For :mod:`builtins` this is just the python interpreter version.

    Args:
        module_name (str): The module to examine.
        version_scraping (VersionScrapingMap | None): Since some modules may store
            their version in other ways, this provides an optional map between module
            names and callables to leverage for extracting that module's version.

    Returns:
        (str | None): The module's version as a string, if any can be found.
    """
    if module_name == "builtins":
        return f"{version_info.major}.{version_info.minor}.{version_info.micro}"

    module_base = module_name.split(".")[0]
    scraper_map: VersionScrapingMap = (
        {} if version_scraping is None else version_scraping
    )

    scraper = (
        scraper_map[module_base]  # noqa: SIM401
        if module_base in scraper_map
        else _scrape_version_attribute
    )
    # mypy struggles with .get even when the fallback is specified,
    # so break it apart and tell Ruff to not worry that we avoid .get
    return scraper(module_base)


def _scrape_version_attribute(module_name: str) -> str | None:
    module = import_module(module_name)
    try:
        return str(module.__version__)
    except AttributeError:
        return None


VersionValidatorType: TypeAlias = (
    Literal["exact", "semantic-minor", "semantic-major", "none"]
    | Callable[[str, str], bool]
)


def validate_version(
    metadata: Metadata,
    validator: VersionValidatorType = "exact",
    version_scraping: VersionScrapingMap | None = None,
) -> None:
    """
    Check whether versioning information in a piece of metadata matches the current
    environment.

    Args:
        metadata (Metadata): The metadata to validate.
        validator ("exact" | Callable[[str, str], bool]): A recognized keyword or a
            callable that takes the current and metadata versions as strings and
            returns a boolean to indicate whether the current version matches the
            metadata reference. Keywords are "exact" (versions must be identical),
            "semantic-minor" (semantic versions (X.Y.Z where all are integers) match
            in the first two digits; all non-semantic versions must match exactly),
            "semantic-major" (semantic versions match in the first digit), and "none"
            (don't compare the versions at all).
        version_scraping (dict[str, Callable[[str], str]] | None): An optional
            dictionary mapping module names to a callable that takes this name and
            returns a version (or None). The default callable imports the module
            string and looks for a `__version__` attribute.

    Raises:
        EnvironmentMismatch: If the module in the metadata cannot be found, or if the
            current and metadata versions do not pass validation.
    """
    if (
        metadata.version is not None
        and metadata.version != ""
        and isinstance(metadata.module, str)
    ):
        try:
            current_version = str(get_version(metadata.module, version_scraping))
        except ModuleNotFoundError as e:
            raise EnvironmentMismatchError(
                f"When unpacking an object, encountered a module {metadata.module}  "
                f"in the metadata that could not be found in the current environment."
            ) from e

        version_validator: VersionValidatorType
        if validator == "exact":
            version_validator = _versions_are_equal
        elif validator == "semantic-minor":
            version_validator = _versions_match_semantic_minor
        elif validator == "semantic-major":
            version_validator = _versions_match_semantic_major
        else:
            version_validator = validator

        if isinstance(version_validator, str):
            if version_validator == "none":
                return
            else:
                raise ValueError(
                    f"Unrecognized validator keyword {version_validator} -- please supply {VersionValidatorType}"
                )
        elif version_validator(current_version, metadata.version):
            return
        raise EnvironmentMismatchError(
            f"{metadata.module} is stored with version {metadata.version}, "
            f"but the current environment has {current_version}. This does not pass "
            f"validation criterion: {version_validator}"
        )


def _versions_are_equal(version: str, reference: str) -> bool:
    return version == reference


def _decompose_semver(version: str) -> tuple[int, int, int] | None:
    match = re.fullmatch(r"(\d+)\.(\d+)\.(\d+)", version)
    if match:
        major, minor, patch = match.groups()
        return int(major), int(minor), int(patch)
    return None


def _versions_match_semantic_minor(version: str, reference: str) -> bool:
    v_parts = _decompose_semver(version)
    r_parts = _decompose_semver(reference)
    if v_parts and r_parts:
        return v_parts[:2] == r_parts[:2]
    return version == reference


def _versions_match_semantic_major(version: str, reference: str) -> bool:
    v_parts = _decompose_semver(version)
    r_parts = _decompose_semver(reference)
    if v_parts and r_parts:
        return v_parts[0] == r_parts[0]
    return version == reference



from __future__ import annotations

from typing import TYPE_CHECKING

import ipytree
import traitlets

from bagofholding.content import Reducible

if TYPE_CHECKING:
    from bagofholding.bag import Bag


class BagTree(ipytree.Tree):  # type: ignore
    # Silence complaints about subclassing `Any` because of upstream issues
    """
    A widget for more convenient bag browsing inside notebooks.
    """

    # pyiron_snippets.import_alarm.ImportAlarm.__call__  is not correctly passing on
    # the hint
    def __init__(self, bag: Bag) -> None:
        super().__init__(multiple_selection=False)
        self.bag = bag
        self.root_path = bag.storage_root
        self.selected_entry = None
        self.observe(self._on_select, names=["selected_nodes"])

        self.root = ipytree.Node("Bag", open=True, disabled=True, icon="shopping-bag")
        self.add_node(self.root)

        for key, value in bag.bag_info.field_items():
            self.root.add_node(
                ipytree.Node(f"{key}: {value}", disabled=True, icon="info")
            )
        self.object = self._create_node(self.root_path)
        self.root.add_node(self.object)

    def _create_node(self, path: str) -> ipytree.Node:
        metadata = self.bag[path]
        try:
            subentries = tuple(k for k in self.bag.open_group(path))
        except TypeError:
            subentries = None

        label_base = path.split("/")[-1]
        truncated_content = metadata.content_type.lstrip(Reducible.__module__)
        label = f"{label_base} ({truncated_content})"

        icon = "file"
        style = "default"
        if metadata.content_type == f"{Reducible.__module__}.{Reducible.__qualname__}":
            icon = "code"
            style = "success"
        elif subentries is not None:
            icon = "folder"
            style = "success" if len(subentries) > 0 else "warning"

        node = ipytree.Node(
            label,
            [],
            opened=False,
            icon=icon,
            open_icon_style=style,
            close_icon_style="danger",
        )
        node.entry_metadata = metadata
        node.entry_subentries = subentries
        node.tree_metadata = {"path": path, "loaded": False}

        # Placeholder for lazy loading subentries
        if subentries and len(subentries) > 0:
            node.add_node(ipytree.Node("Loading...", disabled=True))
            node.observe(self._load_subentries, names=["opened"])

        return node

    def _load_subentries(self, change: traitlets.Bunch) -> None:
        node = change["owner"]

        if node.tree_metadata["loaded"]:
            return

        # Clear placeholder
        node.nodes = []

        if node.entry_metadata:
            for key, value in node.entry_metadata.field_items():
                if value is not None and value != "":
                    node.add_node(
                        ipytree.Node(f"{key}: {value}", disabled=True, icon="info")
                    )

        subentries = node.entry_subentries
        for sub in subentries:
            sub_node = self._create_node(f"{node.tree_metadata['path']}/{sub}")
            node.add_node(sub_node)

        node.tree_metadata["loaded"] = True

    def _on_select(self, change: traitlets.Bunch) -> None:
        if change["new"]:
            self.selected_entry = change["new"][0].tree_metadata["path"]

    def load_selected(self) -> object:
        if self.selected_entry is None:
            raise ValueError("No entry selected")
        print(f"Loading {self.selected_entry}")
        return self.bag.load(self.selected_entry)



import random
from typing import TypeVar, cast

import numpy as np
import pygtrie

ValueType = TypeVar("ValueType")  # python <3.12 compatibility
# def not_allowed_to_have_inline[FunctionGenerics](in_python=3.12):


def decompose_stringtrie(
    trie: pygtrie.StringTrie, null_value: ValueType
) -> tuple[list[str], list[int], list[ValueType]]:
    """
    While tries are a useful representation of path-like tree data, storing the
    individual keys is not space-efficient.

    This function decomposes the trie into a linked-list structure of individual path
    segments and their upstream
    components. Each segment then gets its associated value. Since not all individual
    segments are necessarily valid
    paths in the original tree, we also require a "null" value to keep all components
    the same length.

    Args:
        trie (pygtrie.StringTrie): The trie to decompose.
        null_value (ValueType): What to store in lieu of a value for segments which do
            not correspond to trie paths. Null values must implement a meaningful `==`
            operator with the values of the trie.

    Returns:
        list[str]: The path segments.
        list[int]: Each segment's parent id in the segments list.
        list[ValueType]: The value associated with each segment (taking the null value
            whenever the segment has no associated path in the source trie).
    """
    segments: list[str] = []
    parents: list[int] = []
    values: list[ValueType] = []

    stack = [("", -1)]
    while stack:
        key, parent_idx = stack.pop()
        idx = len(segments)
        segment = key.split(trie._separator)[-1] if key else ""
        segments.append(segment)
        parents.append(parent_idx)
        values.append(
            trie.values(prefix=key, shallow=True)[0]
            if trie.has_key(key)
            else null_value
        )

        prefix = key
        children = set()
        for child_key in trie.keys(prefix=prefix):
            remainder = child_key[len(prefix) :]
            if remainder:
                next_ = remainder.split(trie._separator)[1]
                to_stack = prefix + trie._separator + next_
                children.add(to_stack)

        for child in children:
            stack.append((child, idx))

    return segments, parents, values


def reconstruct_stringtrie(
    segments: list[str],
    parents: list[int],
    values: list[ValueType],
    null_value: ValueType,
    separator: str = "/",
) -> pygtrie.StringTrie:
    """

    While tries are a useful representation of path-like tree data, storing the
    individual keys is not space-efficient.

    This function reconstructs a trie based on a linked-list structure of individual
    path segments and their upstream parent segments.
    Since it is possible that not all segments are intended to be terminal paths in the
    final trie, any segments whose value matches the provided "null" value will not
    actually be added to the resulting trie.

    Args:
        segments (list[str]): The path segments.
        parents (list[int]): The parent ids in the segments list.
        values (list[ValueType]): The corresponding values to assign to the trie paths
            (or do not add if equal to null values).
        null_value (ValueType): The null value against which to compare values to see
            if they should be excluded from the final trie. Null values must implement
            a meaningful `==` operator with elements of the values list.
        separator (str): The path separator.

    Returns:
        (pygtrie.StringTrie): The resulting trie.
    """

    trie = pygtrie.StringTrie(separator=separator)
    keys = [""] * len(segments)

    for i in range(len(segments)):
        if parents[i] == -1:
            keys[i] = ""
        else:
            p = keys[parents[i]]
            keys[i] = p + ("" if p == "" else trie._separator) + segments[i]

    for i, key in enumerate(keys):
        if values[i] != null_value:
            trie[trie._separator + key] = values[i]

    return trie


class Helper:
    """
    Useful tools for building tries on which to test the decomposition of path-like
    tree data.
    """

    @staticmethod
    def compute_softmax_weights(
        n: int, depth_propensity: float, temperature: float = 0.1
    ) -> list[float]:
        pos = np.linspace(0, 1, n)
        scores = (2 * depth_propensity - 1) * pos / temperature
        exp_scores = np.exp(scores - np.max(scores))  # for numerical stability
        return cast(list[float], (exp_scores / exp_scores.sum()).tolist())

    @classmethod
    def generate_paths(
        cls,
        n_paths: int,
        depth_propensity: float = 0.5,
        temperature: float = 0.1,
        seed: int | None = None,
        separator: str = "/",
    ) -> list[str]:
        if seed is not None:
            random.seed(seed)

        paths: list[str] = []
        next_id = 0

        def new_segment() -> str:
            nonlocal next_id
            seg = f"seg{next_id}"
            next_id += 1
            return seg

        frontier: list[list[str]] = [[]]  # list of existing path prefixes

        while len(paths) < n_paths:
            # Choose existing prefix with bias towards depth
            weights = cls.compute_softmax_weights(
                len(frontier), depth_propensity, temperature
            )
            prefix = random.choices(frontier, weights=weights, k=1)[0]

            new_path = prefix + [new_segment()]
            paths.append(separator + separator.join(new_path))

            # Allow extending this path later
            frontier.append(new_path)

        return paths

    @classmethod
    def make_stochastic_trie(
        cls, n_paths: int, depth_propensity: float = 0.5, temperature: float = 0.1
    ) -> tuple[pygtrie.StringTrie, int]:
        trie = pygtrie.StringTrie()
        for i, k in enumerate(
            cls.generate_paths(n_paths, depth_propensity, temperature)
        ):
            trie[k] = i
        null = -1  # real values are >0
        return trie, null



import pathlib
from types import TracebackType
from typing import ClassVar, Literal

import h5py

from bagofholding.exceptions import FileAlreadyOpenError, FileNotOpenError


class HasH5FileContext:
    """
    A mixin class for context management with an :class:`h5py.File` object.
    """

    libver_str: ClassVar[str] = "latest"

    filepath: pathlib.Path
    _file: h5py.File | None
    _context_depth: int

    @property
    def file(self) -> h5py.File:
        if self._file is None:
            raise FileNotOpenError(f"{self.filepath} is not open; use `open` or `with`")
        return self._file

    @file.setter
    def file(self, new_file: h5py.File | None) -> None:
        self._file = new_file

    def open(self, mode: Literal["r", "r+", "w", "w-", "x", "a"]) -> h5py.File:
        if self._file is None:
            self.file = h5py.File(self.filepath, mode, libver=self.libver_str)
            return self.file
        else:
            raise FileAlreadyOpenError(f"The bag at {self.filepath} is already open")

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._context_depth -= 1
        if self._context_depth == 0:
            self.close()

    def close(self) -> None:
        if self._file is not None:
            self.file.close()
            self._file = None

    def __del__(self) -> None:
        self.close()

    @staticmethod
    def maybe_decode(attr: str | bytes) -> str:
        return attr if isinstance(attr, str) else attr.decode("utf-8")



"""
A centralized location for exceptions.

Using custom exceptions allows us to be much more specific in the test suite.
"""

from __future__ import annotations


class BagOfHoldingError(Exception):
    """A base class for raising bagofholding-related exceptions"""


class BagMismatchError(BagOfHoldingError, ValueError): ...


class EnvironmentMismatchError(BagOfHoldingError, ModuleNotFoundError): ...


class FileAlreadyOpenError(BagOfHoldingError): ...


class FileNotOpenError(BagOfHoldingError): ...


class FilepathError(BagOfHoldingError, FileExistsError): ...


class InvalidMetadataError(BagOfHoldingError, ValueError): ...


class ModuleForbiddenError(BagOfHoldingError, ValueError): ...


class NotAGroupError(BagOfHoldingError, TypeError): ...


class NoVersionError(BagOfHoldingError, ValueError): ...


class PickleProtocolError(BagOfHoldingError, ValueError): ...


class StringNotImportableError(BagOfHoldingError): ...



# Designed and intended for use with H5 implementations, which treat arrays very nicely
# but since bags merely follow a protocol, there is nothing intrinsically-h5 here
from __future__ import annotations

from typing import Protocol, TypeAlias

import numpy as np

from bagofholding.content import BespokeItem, Packer, UnpackingArguments
from bagofholding.h5.dtypes import H5DtypeAlias

_INT64_MIN = -9_223_372_036_854_775_808
_UINT64_MAX = 18_446_744_073_709_551_615


def int_overflows(n: int) -> bool:
    """h5py breaks on very extreme integers. Catch them."""
    return n < _INT64_MIN or n > _UINT64_MAX


ArrayType: TypeAlias = np.ndarray[tuple[int, ...], H5DtypeAlias]


class ArrayPacker(Packer, Protocol):
    def pack_array(self, obj: ArrayType, path: str) -> None: ...
    def unpack_array(self, path: str) -> ArrayType: ...


class Array(BespokeItem[ArrayType, ArrayPacker]):
    @classmethod
    def _pack_item(cls, obj: ArrayType, packer: ArrayPacker, path: str) -> None:
        packer.pack_array(obj, path)

    @classmethod
    def unpack(
        cls, packer: ArrayPacker, path: str, unpacking: UnpackingArguments
    ) -> ArrayType:
        return packer.unpack_array(path)
